/*
 * RC Servo driver for VoCore
 *
 * Copyright (C) 2015 Ton Augustin <ton@augustin.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/timer.h>
#include <linux/err.h>
#include <linux/ctype.h>
#include <linux/servos.h>
#include "servos.h"

static struct class *servos_class;

#if 1  ///////////////////// min pulsewidth //////////////////////////////////////////////////////////

static ssize_t min_pulsewidth_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", servo_cdev->min_pulsewidth);
}

static ssize_t min_pulsewidth_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);
	unsigned long mnpw;
	ssize_t ret = -EINVAL;

	ret = kstrtoul(buf, 10, &mnpw);
	if (ret)
		return ret;

	__servo_set_min_pulsewidth(servo_cdev, mnpw);

	return size;
}
static DEVICE_ATTR_RW(min_pulsewidth);
#endif // 1 ///////////////////// end of min pulsewidth //////////////////////////////////////////////////////////


#if 1 ////////////////////// pulsewidth //////////////////////////////////////////////////////////


static ssize_t pulsewidth_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", servo_cdev->pulsewidth);
}

static ssize_t pulsewidth_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);
	unsigned long pulse;
	ssize_t ret = -EINVAL;

	ret = kstrtoul(buf, 10, &pulse);
	if (ret)
		return ret;

	__servo_set_pulsewidth(servo_cdev, pulse);

	return size;
}
static DEVICE_ATTR_RW(pulsewidth);
#endif /////////////////////// end of pulsewidth //////////////////////////////////////////////////////////


#if 1  ///////////////////// max pulsewidth //////////////////////////////////////////////////////////

static ssize_t max_pulsewidth_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", servo_cdev->max_pulsewidth);
}

static ssize_t max_pulsewidth_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);
	unsigned long mxpw;
	ssize_t ret = -EINVAL;

	ret = kstrtoul(buf, 10, &mxpw);
	if (ret)
		return ret;

	__servo_set_max_pulsewidth(servo_cdev, mxpw);

	return size;
}
static DEVICE_ATTR_RW(max_pulsewidth);
#endif // 1 ///////////////////// end of min pulsewidth //////////////////////////////////////////////////////////


#if 1  ///////////////////// timeout //////////////////////////////////////////////////////////

static ssize_t timeout_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	return sprintf(buf, "%u\n", servo_cdev->timeout);
}

static ssize_t timeout_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t size)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);
	unsigned long timeout;
	ssize_t ret = -EINVAL;

	ret = kstrtoul(buf, 10, &timeout);
	if (ret)
		return ret;

	__servo_set_timeout(servo_cdev, timeout);

	return size;
}
static DEVICE_ATTR_RW(timeout);
#endif // 1 ///////////////////// end of timeout //////////////////////////////////////////////////////////


#if 1  ///////////////////// sync //////////////////////////////////////////////////////////
static ssize_t sync_show(struct device *dev,
                            struct device_attribute *attr, char *buf)
{
       struct servo_classdev *servo_cdev = dev_get_drvdata(dev);
	int increase = (servo_cdev->synccurrenttime * servo_cdev->syncdeltaPW) / servo_cdev->synctime;
	
       return sprintf(buf, "time: %d, target: %d active: %d counter: %d deltaPW: %d synctime: %d increase: %d\n", servo_cdev->synctime * MS_TIME_TO_FRAMES, servo_cdev->syncendPW, servo_cdev->syncactive, servo_cdev->synccurrenttime,
			servo_cdev->syncdeltaPW,   servo_cdev->synctime, increase);
	
}

static ssize_t sync_store(struct device *dev,
              struct device_attribute *attr, const char *buf, size_t size)
{
       struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

       unsigned int time = 0, target = 0;
       ssize_t ret = -EINVAL;

       ret = sscanf(buf, "%u %u", &time, &target);

	// Check syntax
       if (ret == 2) {
		// Is this the cancel sync command?
		if((time == 0) && (servo_cdev->syncactive == 1)) {
			servo_cdev->syncactive = 0;						// stop sync
			__servo_set_pulsewidth(servo_cdev, servo_cdev->syncendPW);	// set pulsewidth to target
			return size;								// return
		}	
		if(servo_cdev->syncactive == 1) {
			return -EBUSY;							// Error, already busy	
		}	
		if((target != 0) && (time != 0)) {						// are params valid?
			// Valid command, set up sync, check boundaries
			if (target > servo_cdev->max_pulsewidth)
				target = servo_cdev->max_pulsewidth;
			if (target < servo_cdev->min_pulsewidth)
				target = servo_cdev->min_pulsewidth;

			if (time > MAX_SYNC_TIME) time = MAX_SYNC_TIME;
			if (time < MIN_SYNC_TIME) time = MIN_SYNC_TIME;
			
			servo_cdev->synctime = time / MS_TIME_TO_FRAMES;
			servo_cdev->syncstartPW = servo_cdev->pulsewidth;
			servo_cdev->syncendPW = target;
			servo_cdev->syncdirectionisup = (servo_cdev->syncstartPW > servo_cdev->syncendPW ? 0 : 1);
			servo_cdev->syncdeltaPW =  ( servo_cdev->syncdirectionisup ?  servo_cdev->syncendPW - servo_cdev->syncstartPW : servo_cdev->syncstartPW - servo_cdev->syncendPW );
			servo_cdev->synccurrenttime = 0; 					// init
			servo_cdev->syncactive = 1;						// and go!
			return size;
		}
       } 
       return -EINVAL;
}
static DEVICE_ATTR_RW(sync);
#endif // 1 //////////////// end of sync //////////////////////////////////////////////////////////


static struct attribute *servo_class_attrs[] = {
	&dev_attr_min_pulsewidth.attr,
	&dev_attr_pulsewidth.attr,
	&dev_attr_max_pulsewidth.attr,
	&dev_attr_timeout.attr,
	&dev_attr_sync.attr,
	///&dev_attr_maxold_pulsewidth.attr,
	NULL,
};

static const struct attribute_group servo_group = {
	.attrs = servo_class_attrs,
};

static const struct attribute_group *servo_groups[] = {
	&servo_group,
	NULL,
};

/**
 * servo_classdev_suspend - suspend an servo_classdev.
 * @servo_cdev: the servo_classdev to suspend.
 */
void servo_classdev_suspend(struct servo_classdev *servo_cdev)
{
	servo_cdev->flags |= SERVO_SUSPENDED;
}
EXPORT_SYMBOL_GPL(servo_classdev_suspend);

/**
 * servo_classdev_resume - resume an servo_classdev.
 * @servo_cdev: the servo_classdev to resume.
 */
void servo_classdev_resume(struct servo_classdev *servo_cdev)
{
	servo_cdev->flags &= ~SERVO_SUSPENDED;
}
EXPORT_SYMBOL_GPL(servo_classdev_resume);

static int servo_suspend(struct device *dev)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	if (servo_cdev->flags & SERVO_CORE_SUSPENDRESUME)
		servo_classdev_suspend(servo_cdev);

	return 0;
}

static int servo_resume(struct device *dev)
{
	struct servo_classdev *servo_cdev = dev_get_drvdata(dev);

	if (servo_cdev->flags & SERVO_CORE_SUSPENDRESUME)
		servo_classdev_resume(servo_cdev);

	return 0;
}

static const struct dev_pm_ops servos_class_dev_pm_ops = {
	.suspend        = servo_suspend,
	.resume         = servo_resume,
};

/**
 * servo_classdev_register - register a new object of servo_classdev class.
 * @parent: The device to register.
 * @servo_cdev: the servo_classdev structure for this device.
 */
int servo_classdev_register(struct device *parent, struct servo_classdev *servo_cdev)
{
	servo_cdev->dev = device_create(servos_class, parent, 0, servo_cdev,
				      "%s", servo_cdev->name);
	if (IS_ERR(servo_cdev->dev))
		return PTR_ERR(servo_cdev->dev);

	dev_dbg(parent, "Registered servo device: %s\n",	servo_cdev->name);

	return 0;
}
EXPORT_SYMBOL_GPL(servo_classdev_register);

/**
 * servo_classdev_unregister - unregisters a object of servo_properties class.
 * @servo_cdev: the servo device to unregister
 *
 * Unregisters a previously registered via servo_classdev_register object.
 */
void servo_classdev_unregister(struct servo_classdev *servo_cdev)
{
	device_unregister(servo_cdev->dev);
}
EXPORT_SYMBOL_GPL(servo_classdev_unregister);

static int __init servos_init(void)
{
	servos_class = class_create(THIS_MODULE, "servos");
	if (IS_ERR(servos_class))
		return PTR_ERR(servos_class);
	servos_class->pm = &servos_class_dev_pm_ops;
	servos_class->dev_groups = servo_groups;
	return 0;
}

static void __exit servos_exit(void)
{
	class_destroy(servos_class);
}

subsys_initcall(servos_init);
module_exit(servos_exit);

MODULE_AUTHOR("Ton Augustin");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SERVO Class Interface");
