/*
 * RC Servo driver for VoCore
 *
 * Copyright (C) 2015 Ton Augustin <ton@augustin.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/servos.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/of_gpio.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/err.h>
#include "servos.h" 


static inline int sizeof_gpio_servos_priv(int num_servos)
{
	return sizeof(struct gpio_servos_priv) +
		(sizeof(struct gpio_servo_data) * num_servos);
}

static int create_gpio_servo(const struct gpio_servo *template, 
			struct gpio_servo_data *servo_dat, struct device *parent)
{
	int ret; 

	servo_dat->gpio = -1;

	/* skip servos that aren't available */
	if (!gpio_is_valid(template->gpio)) {
		dev_info(parent, "Skipping unavailable Servo gpio %d (%s)\n", template->gpio, template->name);
		return 0;
	}
	/* skip sleepers */
	if (gpio_cansleep(template->gpio)) {
		dev_info(parent, "No Posh Pins gpio %d (%s)\n",	template->gpio, template->name);
		return 0;
	}

	ret = devm_gpio_request(parent, template->gpio, template->name);
	if (ret < 0)
		return ret;

	servo_dat->cdev.name			= template->name;
	servo_dat->gpio 			= template->gpio;
	servo_dat->active_low 		= template->active_low;
	servo_dat->cdev.pulsewidth 		= SERVO_OFF;
	servo_dat->cdev.timeout		= template->timeout;
	servo_dat->cdev.timeoutactive 	= (servo_dat->cdev.timeout ? 1 : 0 );
	servo_dat->cdev.min_pulsewidth	= template->min_pulsewidth;
	servo_dat->cdev.max_pulsewidth	= template->max_pulsewidth;

	ret = gpio_direction_output(servo_dat->gpio, servo_dat->active_low);
	if (ret < 0)
		return ret;

	ret = servo_classdev_register(parent, &servo_dat->cdev);
	if (ret < 0)
		return ret;

	return 0;
}

static void delete_gpio_servo(struct gpio_servo_data *servo)
{
	if (!gpio_is_valid(servo->gpio))
		return;
	servo_classdev_unregister(&servo->cdev);
}



/// link to the timer.. not so nice like this, but...
struct rt_timer *rt_timer_announce_servos(struct gpio_servos_priv * servos, int count);
int rt_timer_start_servos(struct rt_timer *rt);




/* Code to create from OpenFirmware platform devices */
#ifdef CONFIG_OF_GPIO
static struct gpio_servos_priv *gpio_servos_create_of(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node, *child;
	struct gpio_servos_priv *priv;
	const u32 *prop;	
	int count, ret;

	/* count Servos in this device, so we know how much to allocate */
	count = of_get_available_child_count(np);
	if (!count)
		return ERR_PTR(-ENODEV);

	dev_info(&pdev->dev, "gpio_servos_create_of() found %d servos\n", count);

	
	for_each_available_child_of_node(np, child)
		if (of_get_gpio(child, 0) == -EPROBE_DEFER)
			return ERR_PTR(-EPROBE_DEFER);

	priv = devm_kzalloc(&pdev->dev, sizeof_gpio_servos_priv(count),
			GFP_KERNEL);
	if (!priv)
		return ERR_PTR(-ENOMEM);
	
	/// connect with timer, store ptr to it
	priv->rtimer = rt_timer_announce_servos(priv, count);	
		
	for_each_available_child_of_node(np, child) {
		struct gpio_servo servo = {};
		enum of_gpio_flags flags;

		servo.gpio = of_get_gpio_flags(child, 0, &flags);
		servo.active_low = flags & OF_GPIO_ACTIVE_LOW;
		servo.name = of_get_property(child, "label", NULL) ? : child->name;

		prop = of_get_property(child, "timeout", NULL);
		if (!prop) {
			servo.timeout = 0;
		} else {
			servo.timeout = *prop;
		}

		prop = of_get_property(child, "min_pulsewidth", NULL);
		if (!prop) {
			servo.min_pulsewidth = SERVO_MIN;
		} else {
			servo.min_pulsewidth = *prop;
			if (servo.min_pulsewidth < SERVO_MIN)
				servo.min_pulsewidth = SERVO_MIN;
		}

		prop = of_get_property(child, "max_pulsewidth", NULL);
		if (!prop) {
			servo.max_pulsewidth = SERVO_MAX;
		} else {
			servo.max_pulsewidth = *prop;
			if (servo.max_pulsewidth > SERVO_MAX)
				servo.max_pulsewidth = SERVO_MAX;
		}
		
		ret = create_gpio_servo(&servo, &priv->servo[priv->num_servos++], &pdev->dev );
		if (ret < 0) {
			of_node_put(child);
			goto err;
		}
	}
	rt_timer_start_servos(priv->rtimer);
	return priv;

err:
	for (count = priv->num_servos - 2; count >= 0; count--)
		delete_gpio_servo(&priv->servo[count]);
	return ERR_PTR(-ENODEV);
}

static const struct of_device_id of_gpio_servos_match[] = {
	{ .compatible = "servocore", },
	{},
};
#else /* CONFIG_OF_GPIO */
static struct gpio_servos_priv *gpio_servos_create_of(struct platform_device *pdev)
{
	return ERR_PTR(-ENODEV);
}
#endif /* CONFIG_OF_GPIO */


static int gpio_servo_probe(struct platform_device *pdev)
{
	struct gpio_servo_platform_data *pdata = dev_get_platdata(&pdev->dev);
	struct gpio_servos_priv *priv;
	int i, ret = 0;


	if (pdata && pdata->num_servos) {
		priv = devm_kzalloc(&pdev->dev,
				sizeof_gpio_servos_priv(pdata->num_servos),
					GFP_KERNEL);
		if (!priv)
			return -ENOMEM;

		priv->num_servos = pdata->num_servos;
		for (i = 0; i < priv->num_servos; i++) {
			ret = create_gpio_servo(&pdata->servo[i],
					      &priv->servo[i],
					      &pdev->dev );
			if (ret < 0) {
				/* On failure: unwind the servo creations */
				for (i = i - 1; i >= 0; i--)
					delete_gpio_servo(&priv->servo[i]);
				return ret;
			}
		}
	} else {
		priv = gpio_servos_create_of(pdev);
		if (IS_ERR(priv))
			return PTR_ERR(priv);
	}

	platform_set_drvdata(pdev, priv);

	return 0;
}

static int gpio_servo_remove(struct platform_device *pdev)
{
	struct gpio_servos_priv *priv = platform_get_drvdata(pdev);
	int i;

	for (i = 0; i < priv->num_servos; i++)
		delete_gpio_servo(&priv->servo[i]);

	return 0;
}

static struct platform_driver gpio_servo_driver = {
	.probe		= gpio_servo_probe,
	.remove	= gpio_servo_remove,
	.driver	= {
		.name	= "servocore",
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(of_gpio_servos_match),
	},
};

module_platform_driver(gpio_servo_driver);

MODULE_AUTHOR("Ton Augustin <ton@augustin.com>");
MODULE_DESCRIPTION("RC Servo driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:servocore");
