/*
 * timer.c
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 *
 * Copyright (C) 2013 John Crispin <blogic@openwrt.org>
 *
 * Modified for Servocore 2015 by Ton Augustin <ton@augustin.com>
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/of_gpio.h>
#include <linux/clk.h>
#include <linux/servos.h>
#include <linux/slab.h>

#include <asm/mach-ralink/ralink_regs.h>

#define TIMER_REG_TMRSTAT          0x00
#define TIMER_REG_TMR0LOAD         0x10
#define TIMER_REG_TMR0CTL          0x18

#define TMRSTAT_TMR0INT            BIT(0)

#define TMR0CTL_ENABLE             BIT(7)
#define TMR0CTL_MODE_PERIODIC      BIT(4)
// was 2
#define TMR0CTL_PRESCALER          12
#define TMR0CTL_PRESCALE_VAL       (0xf - TMR0CTL_PRESCALER)
#define TMR0CTL_PRESCALE_DIV       (65536 / BIT(TMR0CTL_PRESCALER))



struct rt_timer {
       struct device        *dev;
       void __iomem         *membase;
       int                  irq;
	int			announced;

       unsigned long        timer_freq;
       unsigned long        timer_div;

	struct gpio_servos_priv *servos;

	int			current_servo;
	int			frametime;
	int			cumulativepulsetime;
	
};

struct rt_timer *rtforservos = NULL;

static inline void rt_timer_w32(struct rt_timer *rt, u8 reg, u32 val)
{
       __raw_writel(val, rt->membase + reg);
}

static inline u32 rt_timer_r32(struct rt_timer *rt, u8 reg)
{
       return __raw_readl(rt->membase + reg);
}

static irqreturn_t rt_timer_irq(int irq, void *_rt)
{
       struct rt_timer *rt =  (struct rt_timer *) _rt;
	struct gpio_servos_priv *servos = rt->servos;
       unsigned int val;
	int level, increase;
	
	// switch off previous servo-pulse, no switch off after frametime
	if(rt->current_servo != servos->num_servos) {		 
		level = 0; // turn off
		if (servos->servo[rt->current_servo].active_low)
			level = !level;
		gpio_set_value(servos->servo[rt->current_servo].gpio, level);	
	}

findservo:	
	// this part  finds a next servo that actually pulses (!= SERVO_OFF). If none, it ends in a frame-long interrupt
	// we can goto after clearing the pulse because it was never set
	rt->current_servo++;

	if(rt->current_servo ==  servos->num_servos) {		// We did all servos, so now fill up the framespace
		// calculate framespace = frameperiod - pulsestime and set val to framespace
		val = rt->frametime - rt->cumulativepulsetime;
		rt->cumulativepulsetime = 0; 	// reset pulstime
	} else {

		if(rt->current_servo ==  servos->num_servos + 1) { // round robin after frametimer
			rt->current_servo = 0;
		}
		// check if in sync, if so, adjust pulsewidth
		if(servos->servo[rt->current_servo].cdev.syncactive == 1) {
			// Are we at the end? If so end it all.
			if(++servos->servo[rt->current_servo].cdev.synccurrenttime == servos->servo[rt->current_servo].cdev.synctime) {		
				servos->servo[rt->current_servo].cdev.pulsewidth = servos->servo[rt->current_servo].cdev.syncendPW;
				servos->servo[rt->current_servo].cdev.syncactive = 0;
			} else {
				// not yet at the end, calculate new position and check boundaries
				increase = (servos->servo[rt->current_servo].cdev.synccurrenttime * servos->servo[rt->current_servo].cdev.syncdeltaPW) / servos->servo[rt->current_servo].cdev.synctime;
				if(servos->servo[rt->current_servo].cdev.syncdirectionisup == 1) {
					servos->servo[rt->current_servo].cdev.pulsewidth = servos->servo[rt->current_servo].cdev.syncstartPW + increase;
					if( servos->servo[rt->current_servo].cdev.pulsewidth >= servos->servo[rt->current_servo].cdev.syncendPW ) { 	// Are we ready? if so end
						servos->servo[rt->current_servo].cdev.pulsewidth = servos->servo[rt->current_servo].cdev.syncendPW;
						servos->servo[rt->current_servo].cdev.syncactive = 0;
					}
				} else {
					servos->servo[rt->current_servo].cdev.pulsewidth = servos->servo[rt->current_servo].cdev.syncstartPW - increase;
					if( servos->servo[rt->current_servo].cdev.pulsewidth <= servos->servo[rt->current_servo].cdev.syncendPW ) { 	// Are we ready? if so end
						servos->servo[rt->current_servo].cdev.pulsewidth = servos->servo[rt->current_servo].cdev.syncendPW;
						servos->servo[rt->current_servo].cdev.syncactive = 0;
					}
				}
			}
			// and set timeout if needed
			if(servos->servo[rt->current_servo].cdev.timeoutactive) servos->servo[rt->current_servo].cdev.timeoutcnt = servos->servo[rt->current_servo].cdev.timeout;
		} 
		
		val = servos->servo[rt->current_servo].cdev.pulsewidth;
		// when SERVO_OFF, skip this one and go to the next
		// should this one be after the timeout??????????
		if (val == SERVO_OFF) goto findservo;

		// when timeout is active, check if pulse is needed
		if(servos->servo[rt->current_servo].cdev.timeoutactive) {
			// when cnt == 0, old stuff, no pulse.  < 1 to avoid missing a -- and never end exactly on 0 
			if(servos->servo[rt->current_servo].cdev.timeoutcnt < 1) goto findservo;
			servos->servo[rt->current_servo].cdev.timeoutcnt--;
		}
		
		rt->cumulativepulsetime += val;

		level = 1;
		if (servos->servo[rt->current_servo].active_low)
			level = !level;
		gpio_set_value(servos->servo[rt->current_servo].gpio, level);	
	}
	
	val = val * MS_TIME_TO_FRAMES; // stay under 0xFFFE, 16 bits register!

	rt_timer_w32(rt, TIMER_REG_TMR0LOAD, val + 1);

       rt_timer_w32(rt, TIMER_REG_TMRSTAT, TMRSTAT_TMR0INT);

       return IRQ_HANDLED;
}


static int rt_timer_request(struct rt_timer *rt)
{
       int err = devm_request_irq(rt->dev, rt->irq, rt_timer_irq,
                                   IRQF_DISABLED, dev_name(rt->dev), rt);
       if (err) {
              dev_err(rt->dev, "failed to request irq\n");
       } else {
              u32 t = TMR0CTL_MODE_PERIODIC | TMR0CTL_PRESCALE_VAL;
              rt_timer_w32(rt, TIMER_REG_TMR0CTL, t);
       }
       return err;
}

static void rt_timer_free(struct rt_timer *rt)
{
       free_irq(rt->irq, rt);
}

static int rt_timer_config(struct rt_timer *rt, unsigned long divisor)
{
       if (rt->timer_freq < divisor)
              rt->timer_div = rt->timer_freq;
       else
              rt->timer_div = divisor;

       return 0;
}

static int rt_timer_enable(struct rt_timer *rt)
{
       u32 t;

       rt_timer_w32(rt, TIMER_REG_TMR0LOAD, 65000 );

       ///PPP t = rt_timer_r32(rt, TIMER_REG_TMR0CTL);
	///PPP be sure prescale is ok. that was a big big problem, prescaler now WORKS!
	t = TMR0CTL_MODE_PERIODIC | TMR0CTL_PRESCALE_VAL;
       t |= TMR0CTL_ENABLE;
       rt_timer_w32(rt, TIMER_REG_TMR0CTL, t);

       return 0;
}

static void rt_timer_disable(struct rt_timer *rt)
{
       u32 t;

       t = rt_timer_r32(rt, TIMER_REG_TMR0CTL);
       t &= ~TMR0CTL_ENABLE;
       rt_timer_w32(rt, TIMER_REG_TMR0CTL, t);
}


struct rt_timer *rt_timer_announce_servos(struct gpio_servos_priv * servos, int count)
{
	struct rt_timer *rt = rtforservos;

	// receive the pointer to array of servos 
	rt->servos = servos;
	if(count > 0) {
		rt->announced = 1;
		rt->frametime = FRAMETIME;	
	}
	return rt;
}
EXPORT_SYMBOL(rt_timer_announce_servos);


int rt_timer_start_servos(struct rt_timer *rt)
{
	// if not announced, error
	if(rt->announced != 1) return -1; /// errorcode
	
	// init timer, start the loop
	rt->current_servo = 0;
	rt->cumulativepulsetime = 0;
	rt_timer_enable(rt);
	return 0;
}	
EXPORT_SYMBOL(rt_timer_start_servos);


static int rt_timer_probe(struct platform_device *pdev)
{
       struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
       const __be32 *divisor;
       struct rt_timer *rt;
       struct clk *clk;
       int ret;


       rt = devm_kzalloc(&pdev->dev, sizeof(*rt), GFP_KERNEL);
       if (!rt) {
              dev_err(&pdev->dev, "failed to allocate memory\n");
              return -ENOMEM;
       }

       rt->irq = platform_get_irq(pdev, 0);
       if (!rt->irq) {
              dev_err(&pdev->dev, "failed to load irq\n");
              return -ENOENT;
       }

	// we did not get servos yet, so do not start.
	rt->announced = 0;
	
       rt->membase = devm_ioremap_resource(&pdev->dev, res);
       if (IS_ERR(rt->membase))
              return PTR_ERR(rt->membase);

       clk = devm_clk_get(&pdev->dev, NULL);
       if (IS_ERR(clk)) {
              dev_err(&pdev->dev, "failed get clock rate\n");
              return PTR_ERR(clk);
       }

       rt->timer_freq = clk_get_rate(clk) / TMR0CTL_PRESCALE_DIV;
       if (!rt->timer_freq)
              return -EINVAL;

       //rt->duty_cycle = 100;
       rt->dev = &pdev->dev;
       platform_set_drvdata(pdev, rt);

	/// QD preserve the pointer, hand it out at announce.. yeah yeah
	rtforservos = rt;
	
       ret = rt_timer_request(rt);
       if (ret)
              return ret;

       divisor = of_get_property(pdev->dev.of_node, "ralink,divisor", NULL);
       if (divisor)
              rt_timer_config(rt, be32_to_cpu(*divisor));
       else
              rt_timer_config(rt, 200);

       dev_info(&pdev->dev, "mmmmaximum frequency is %luHz\n", rt->timer_freq);
       dev_info(&pdev->dev, "divisor           is %lu\n", rt->timer_div);

       return 0;
}

static int rt_timer_remove(struct platform_device *pdev)
{
       struct rt_timer *rt = platform_get_drvdata(pdev);

       rt_timer_disable(rt);
       rt_timer_free(rt);

       return 0;
}

static const struct of_device_id rt_timer_match[] = {
       { .compatible = "ralink,rt2880-timer" },
       {},
};
MODULE_DEVICE_TABLE(of, rt_timer_match);

static struct platform_driver rt_timer_driver = {
       .probe = rt_timer_probe,
       .remove = rt_timer_remove,
       .driver = {
              .name         = "rt-timer",
              .owner          = THIS_MODULE,
              .of_match_table      = rt_timer_match
       },
};

module_platform_driver(rt_timer_driver);

MODULE_DESCRIPTION("Ralink RT2880 timer / servocore");
MODULE_AUTHOR("John Crispin / Ton Augustin");
MODULE_LICENSE("GPL");
